#include "MbufModule.h"

namespace mbuf
{
  bool MbufModule::initialized = false;

  /// @brief this method initialize Mbuf and pool.
  /// @param areYouMbufWriter true if you are creator of this Mbuf, else false.
  /// @param mbufBlockNumber number of blocks of this Mbuf. It's default value
  /// is 20.
  /// @param mbufBlockSize size of a block of Mbuf. It's default value is 1048576.
  void MbufModule::init(bool areYouMbufWriter, uint32_t mbufBlockNumber, uint32_t mbufBlockSize)
  {
    if (!initialized)
    {
      if (!ara::rm::RegisterHisiResource())
      {
        this->log_.LogError() << "Register rm failed!, please check whether rm_center process exists";
      }

      // Init Mbuf
      BuffCfg cfg;
      memset_s(&cfg, sizeof(BuffCfg), 0x0, sizeof(BuffCfg));
      const auto bufInitRet = halBuffInit(&cfg);
      if (bufInitRet != DRV_ERROR_NONE)
      {
        this->log_.LogError() << "Init mbuf error, error code is: " << bufInitRet;
        return;
      }
      else
      {
        this->log_.LogInfo() << "Init mbuf completed without any error";
      }
      initialized = true;
    }
    if (areYouMbufWriter)
    {
      // Create Pool
      constexpr uint32_t alignSize = 32U;
      mpAttr attr;
      attr.devid = 0;
      attr.mGroupId = 0;
      attr.blkSize = ((mbufBlockSize + alignSize) - 1U) & (~(alignSize - 1U));
      attr.blkNum = mbufBlockNumber;
      attr.align = alignSize;
      attr.hugePageFlag = 1U | BUFF_SP_DVPP;
      // attr.hugePageFlag = BUFF_SP_HUGEPAGE_ONLY;
      const auto poolCreateRet = halBuffCreatePool(&attr, &this->handle);
      if (poolCreateRet != DRV_ERROR_NONE)
      {
        this->log_.LogError() << "Create pool error, error code is: " << poolCreateRet;
      }
      else
      {
        this->log_.LogDebug() << "Create pool completed without any error";
      }
    }
  }

  /// @brief this method fill the mbuf with given data and return pointer of it.
  /// @param data a vector of uint8_t which contains your data. You should cast
  /// you data to it yourself!
  /// @param mbufSize is the size of Mbuf which is filled with this method.
  /// @return pointer of filled Mbuf with your data.
  Mbuf *MbufModule::fillMbuf(std::vector<uint8_t> data, uint64_t *mbufSize)
  {
    // Create Mbuf
    const auto bufCreateRet = halMbufAllocByPool(this->handle, &this->pMbuf);
    if (bufCreateRet != DRV_ERROR_NONE)
    {
      this->log_.LogError() << "Allocate Mbuf error, error code is: " << bufCreateRet;
    }
    else
    {
      this->log_.LogDebug() << "Allocate Mbuf completed without any error";
    }

    // Fill Mbuf
    uintptr_t dst;
    uint64_t bufSize = 0U;
    void *pData = nullptr;
    const auto bufPtrRet = halMbufGetDataPtr(this->pMbuf, &pData, &bufSize);
    if (bufPtrRet != DRV_ERROR_NONE)
    {
      this->log_.LogError() << "Get Mbuf data pointer error, error code is: " << bufPtrRet;
    }
    else
    {
      this->log_.LogDebug() << "Get Mbuf data pointer completed without any error";
    }
    dst = reinterpret_cast<uintptr_t>(pData);
    memcpy_s(reinterpret_cast<uint8_t *>(dst), static_cast<uint64_t>(bufSize), data.data(), static_cast<uint64_t>(data.size()));
    const auto bufSetRet = halMbufSetDataLen(this->pMbuf, data.size());
    if (bufSetRet != DRV_ERROR_NONE)
    {
      this->log_.LogError() << "Set Mbuf data length error, error code is: " << bufSetRet;
    }
    else
    {
      this->log_.LogDebug() << "Set Mbuf data length completed without any error";
    }

    // Fill mbufSize
    const auto bufLenRet = halMbufGetDataLen(this->pMbuf, mbufSize);
    if (bufLenRet != DRV_ERROR_NONE)
    {
      this->log_.LogError() << "Get Mbuf data Length error, error code is: " << bufLenRet;
    }
    else
    {
      this->log_.LogDebug() << "Get Mbuf data length completed without any error";
    }

    return this->pMbuf;
  }

  Mbuf *MbufModule::fillMbuf(void *data, uint64_t *mbufSize, uint64_t dataSize)
  {
    // Create Mbuf
    const auto bufCreateRet = halMbufAllocByPool(this->handle, &this->pMbuf);
    if (bufCreateRet != DRV_ERROR_NONE)
    {
      this->log_.LogDebug() << "Allocate Mbuf error, error code is: " << bufCreateRet;
    }
    else
    {
      this->log_.LogDebug() << "Allocate Mbuf completed without any error";
    }

    // Fill Mbuf
    uint64_t bufSize = 0U;
    void *pData = nullptr;
    const auto bufPtrRet = halMbufGetDataPtr(this->pMbuf, &pData, &bufSize);
    if (bufPtrRet != DRV_ERROR_NONE)
    {
      this->log_.LogError() << "Get Mbuf data pointer error, error code is: " << bufPtrRet;
    }
    else
    {
      this->log_.LogDebug() << "Get Mbuf data pointer completed without any error";
    }
    memcpy(pData, data, dataSize);
    // pData = data;
    const auto bufSetRet = halMbufSetDataLen(this->pMbuf, dataSize);
    if (bufSetRet != DRV_ERROR_NONE)
    {
      this->log_.LogError() << "Set Mbuf data length error, error code is: " << bufSetRet;
    }
    else
    {
      this->log_.LogDebug() << "Set Mbuf data length completed without any error";
    }

    // Fill mbufSize
    const auto bufLenRet = halMbufGetDataLen(this->pMbuf, mbufSize);
    if (bufLenRet != DRV_ERROR_NONE)
    {
      this->log_.LogError() << "Get Mbuf data Length error, error code is: " << bufLenRet;
    }
    else
    {
      this->log_.LogDebug() << "Get Mbuf data length completed without any error";
    }

    return this->pMbuf;
  }

  /// @brief this method retrieve the data of given Mbuf and return it. This
  /// method free the Mbuf at the end.
  /// @param buffer pointer of Mbut that contains your data. Send it by reference
  /// since this method should free it.
  /// @param size size of your actual data.
  /// @return a pointer to the beginning of your actual data. You can iterate it
  /// since you have its size.
  uint8_t *MbufModule::getMbufData(Mbuf **buffer, uint64_t *size)
  {
    uint8_t *pData;
    auto ret = halMbufGetDataPtr(*buffer, reinterpret_cast<void **>(&pData), size);
    if (ret != 0)
    {
      this->log_.LogError() << "Get Mbuf data pointer error, error code is: " << ret;
    }
    else
    {
      this->log_.LogDebug() << "Get Mbuf data pointer completed without any error";
    }
    auto freeRet = halMbufFree(*buffer);
    if (freeRet != 0)
    {
      this->log_.LogError() << "free Mbuf error, error code is: " << freeRet;
    }
    else
    {
      this->log_.LogDebug() << "free Mbuf completed without any error";
    }
    return pData;
  }
} // namespace mbuf
